/**
 * Copyright (C) 2012 JBoss Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jbpm.formModeler.core.processing.fieldHandlers;

import org.apache.commons.lang.StringUtils;
import org.jbpm.formModeler.api.model.DataHolder;
import org.jbpm.formModeler.api.model.Field;
import org.jbpm.formModeler.api.model.Form;
import org.jbpm.formModeler.core.processing.*;
import org.jbpm.formModeler.core.processing.fieldHandlers.subform.checkers.SubformChecker;
import org.jbpm.formModeler.core.processing.formRendering.FormErrorMessageBuilder;
import org.jbpm.formModeler.core.processing.formStatus.FormStatus;
import org.jbpm.formModeler.core.rendering.SubformFinderService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.PostConstruct;
import javax.enterprise.inject.Instance;
import javax.inject.Inject;
import javax.inject.Named;
import java.util.*;

@Named("org.jbpm.formModeler.core.processing.fieldHandlers.SubformFieldHandler")
public class SubformFieldHandler extends PersistentFieldHandler {
    private static transient Logger log = LoggerFactory.getLogger(SubformFieldHandler.class);

    @Inject
    private Instance<SubformChecker> checkers;
    private Set<SubformChecker> subformCheckers;

    @Inject
    private SubformFinderService subformFinderService;

    @Inject
    private FormErrorMessageBuilder formErrorMessageBuilder;

    private String pageToIncludeForRendering = "/formModeler/fieldHandlers/Subform/input.jsp";
    private String pageToIncludeForDisplaying = "/formModeler/fieldHandlers/Subform/show.jsp";
    private String pageToIncludeForSearching = "/formModeler/fieldHandlers/Subform/search.jsp";

    private static int maxDepth = 2;


    @PostConstruct
    public void prepare() {
        subformCheckers = new TreeSet<SubformChecker>(new Comparator<SubformChecker>() {
            @Override
            public int compare(SubformChecker o1, SubformChecker o2) {
                return o1.getPriority() - o2.getPriority();
            }
        });
        for (SubformChecker p : checkers) {
            subformCheckers.add(p);
        }
    }

    /**
     * Determine the list of class types this field can generate. That is, normally,
     * a field can generate multiple outputs (an input text can generate Strings,
     * Integers, ...)
     *
     * @return the set of class types that can be generated by this handler.
     */
    public String[] getCompatibleClassNames() {
        return new String[]{Object.class.getName()};
    }


    public Object getValue(Field field, String inputName, Map parametersMap, Map filesMap, String desiredClassName, Object previousValue) throws Exception {
        Form form = getEnterDataForm(inputName, field);
        if (!checkSubformDepthAllowed(form, inputName)) return null;
        getFormProcessor().setValues(form, inputName, parametersMap, filesMap);
        FormStatusData status = getFormProcessor().read(form, inputName);
        if (status.isValid()) {
            // Check if form status is empty & if the object already exists to avoid null objects creation.
            if (status.isEmpty()) return null;
            Map m = getFormProcessor().getMapRepresentationToPersist(form, inputName);
            return m;
        } else {
            throw new IllegalArgumentException("Subform status is invalid.");
        }
    }

    @Override
    public Object persist(Field field, String inputName) throws Exception {
        Form form = getEnterDataForm(inputName, field);
        Map representation = getFormProcessor().getMapRepresentationToPersist(form, inputName);

        DataHolder holder = form.getHolders().iterator().next();

        FormStatus fs = getFormStatusManager().getFormStatus(form, inputName);
        Object locadedObject = fs.getLoadedObject(holder.getUniqeId());

        return getFormProcessor().persistFormHolder(form, inputName, representation, holder, locadedObject);
    }

    @Override
    public Object getStatusValue(Field field, String inputName, Object value) {
        if (value == null) return Collections.EMPTY_MAP;
        Form form = getEnterDataForm(inputName, field);
        DataHolder holder = form.getHolders().iterator().next();

        Map<String, Object> inputData = new HashMap();

        if (!StringUtils.isEmpty(holder.getInputId())) inputData.put(holder.getInputId(), value);

        Map<String, Object> outputData = new HashMap();

        if (!StringUtils.isEmpty(holder.getInputId())) outputData.put(holder.getOuputId(), value);

        Map<String, Object> loadedObjects = new HashMap();

        Map result = null;
        try {
            result = getFormProcessor().createFieldContextValueFromHolder(form, inputName, inputData, outputData, loadedObjects, holder);
        } catch (Exception e) {
            log.error("Error getting status value for field: " + inputName, e);
        }

        getFormProcessor().read(form, inputName, result, loadedObjects);

        return result;
    }

    public Map getParamValue(String inputName, Object value, String pattern) {
        if (value == null)
            return Collections.EMPTY_MAP;

        FormNamespaceData data = getNamespaceManager().getNamespace(inputName);
        Field parentField = data.getForm().getField(data.getFieldNameInParent());
        Form childForm = getEnterDataForm(inputName, parentField);

        Map result = new HashMap();
        DataHolder holder = childForm.getHolders().iterator().next();
        for (Iterator it = childForm.getFormFields().iterator(); it.hasNext();) {
            Field childField = (Field) it.next();
            String bindingExpression = StringUtils.defaultIfEmpty(childField.getInputBinding(), childField.getOutputBinding());
            if (!holder.isAssignableForField(childField)) continue;
            try {
                Object val = holder.readFromBindingExperssion(value, bindingExpression);
                FieldHandler fieldManager = getFieldHandlersManager().getHandler(childField.getFieldType());
                Map childrenMap = fieldManager.getParamValue(childField.getFieldName(), val, childField.getFieldPattern());
                if (childrenMap != null) result.putAll(childrenMap);
            } catch (Exception e) {
                log.warn("Error reading value from field '" + childField.getFieldName() + "': ", e);
            }
        }

        return result;
    }


    public List getWrongChildFieldErrors(String namespace, Field field) {
        return formErrorMessageBuilder.getWrongFormErrors(namespace, getEnterDataForm(namespace, field));
    }

    /**
     * When rendering a form, if field is handled by this handler, determine the
     * page that renders the displaying of the value
     *
     * @return a page to include
     */
    public String getPageToIncludeForDisplaying() {
        return pageToIncludeForDisplaying;
    }

    public boolean isEmpty(Object value) {
        return value == null || "".equals(value);
    }

    public void setPageToIncludeForDisplaying(String pageToIncludeForDisplaying) {
        this.pageToIncludeForDisplaying = pageToIncludeForDisplaying;
    }

    public boolean acceptsPropertyName(String propName) {
        return true;
    }

    /**
     * When rendering a form, if field is handled by this handler, determine the
     * page that renders the input(s)
     *
     * @return a page to include
     */
    public String getPageToIncludeForRendering() {
        return pageToIncludeForRendering;
    }

    public void setPageToIncludeForRendering(String pageToIncludeForRendering) {
        this.pageToIncludeForRendering = pageToIncludeForRendering;
    }

    public String getPageToIncludeForSearching() {
        return pageToIncludeForSearching;
    }

    public void setPageToIncludeForSearching(String pageToIncludeForSearching) {
        this.pageToIncludeForSearching = pageToIncludeForSearching;
    }

    protected Form getEnterDataForm(String namespace, Field field) {
        String formName = field.getDefaultSubform();

        return getForm(formName, namespace);
    }

    private Form getForm(String formPath, String namespace) {
        return subformFinderService.getFormFromPath(formPath, namespace);
    }

    public static boolean checkSubformDepthAllowed(Form form, String namesapce) {
        StringTokenizer token = new StringTokenizer(namesapce, FormProcessor.NAMESPACE_SEPARATOR, false);
        String _id = form.getId().toString();
        int count = 0;
        while (token.hasMoreElements()) {
            String idToCompare = (String) token.nextElement();
            if (idToCompare.equals(_id)) {
                count++;
                if (count >= maxDepth) {
                    return false;
                }
            }

        }
        return true;

    }

    public Set<SubformChecker> getSubformCheckers() {
        return subformCheckers;
    }
}
